/**
 * Copyright 2007 Nathanael Cottin
 */
package sds;

import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.HashSet;

import sds.data.Reference;
import sds.more.Codec;

/**
 * <p>
 * Internal message delivery functionality
 * </p>
 * 
 * @see DistributedObject#deliver(sds.data.Message, boolean, Reference...)
 * @author ncottin
 */
final class Delivery extends Thread {

	private final SecureMessageFormat	smf;
	private final Reference[]			recipients;
	private final Collection<Reference>	invalidRecipients;

	public Delivery(SecureMessageFormat smf, Reference... recipients) {
		this.smf = smf;
		this.recipients = recipients;
		invalidRecipients = new HashSet<Reference>(recipients.length);
	}

	@Override
	public void run() {
		if ((recipients == null) || (smf == null)) {
			return;
		}

		SignedData tbsData = smf.getTbsData();
		if (tbsData == null) {
			return;
		}

		MessageData msgData = tbsData.getTbsData();
		if (msgData == null) {
			return;
		}

		for (Reference dest : recipients) {
			if ((dest != null) && dest.isInvokable()) {
				msgData.setRecipient(dest);
				try {
					send(smf, dest);
				}
				catch (Exception e) {
					invalidRecipients.add(dest);
				}
			}
		}
	}

	private void send(SecureMessageFormat smf, Reference dest)
		throws UnknownHostException,
			IOException {
		Socket sock = new Socket(dest.getAddress(), dest.getPort());
		sock.shutdownInput();
		OutputStream out = sock.getOutputStream();
		Codec.encode(smf, out);
		out.flush();
		out.close();
		sock.close();
	}

	public Collection<Reference> getInvalidRecipients() {
		return invalidRecipients;
	}
}
