/**
 * Copyright 2007 Nathanael Cottin
 */
package sds;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;

import sds.data.Message;
import sds.data.Reference;
import sds.data.TransactionInfo;
import sds.error.SdsException;
import sds.more.SimpleRandom;

/**
 * <p>
 * Defines generic distributed objects manipulated by SDS
 * </p>
 * 
 * @author ncottin
 */
public abstract class DistributedObject {

	private static final long				ZERO	= 0;
	private final Map<Long, ResponseInfo>	expectedResponses;
	private Reference						manager;
	private Reference						ref;

	/**
	 * <p>
	 * Creates a non-initialized distributed object along with no associated
	 * manager and a default produced reference
	 * </p>
	 */
	public DistributedObject() {
		expectedResponses = new Hashtable<Long, ResponseInfo>();
	}

	/**
	 * <p>
	 * Used by servants to indicate that a received message data contains the
	 * given transaction identifier. This operation contributes to unblock
	 * waiting requests
	 * </p>
	 * 
	 * @param id
	 */
	final void countResponse(Long id) {
		if (id == null) {
			return;
		}

		ResponseInfo value = expectedResponses.get(id);
		if (value == null) {
			return;
		}

		AtomicLong respId = value.getCount();
		respId.decrementAndGet();
		if (respId.longValue() == ZERO) {
			// No more expected response
			expectedResponses.remove(id);
			value.unlock();
		}
	}

	public final ResponseInfo getExpectedResponseInfo(Long id) {
		return (id == null)
			? null
			: expectedResponses.get(id);
	}

	/**
	 * <p>
	 * To be used by a daemon to set the manager associated with this
	 * distributed object
	 * </p>
	 * 
	 * @param manager the manager which called the local daemon to activate this
	 * distributed object. Setting <tt>null</tt> indicates that ths
	 * distributed object is not associated with any manager
	 */
	public final void setManager(Reference manager) {
		this.manager = manager;
	}

	/**
	 * <p>
	 * Gets the reference of the manager which sent this distributed object to
	 * the local daemon
	 * </p>
	 * 
	 * @return the manager's reference or <tt>null</tt> if this object is not
	 * associated with a manager
	 */
	public final Reference getManager() {
		return manager;
	}

	public final void setDefaultReference() {
		ref = createReference();
	}

	public static final Reference createReference() {
		Reference ref = new Reference();
		ref.setNonce(SimpleRandom.nextLong());
		return ref;
	}

	public final void setReference(Reference reference) {
		ref = reference;
	}

	/**
	 * <p>
	 * Gets the reference of this distributed object
	 * </p>
	 * 
	 * @return this distributed object's fully initialized reference
	 */
	public final Reference getReference() {
		return ref;
	}

	/**
	 * <p>
	 * Indicated whether this distributed object is a server (or an
	 * intermediary) or a client
	 * </p>
	 * 
	 * @return <tt>true</tt> only if this distributed object is able to listen
	 * to incoming requests (i.e. this distributed object is server-oriented)
	 */
	public abstract boolean runsAsServer();

	/**
	 * <p>
	 * Executes this distributed object as a server. This operation must be
	 * non-blocking. A default reference is created if not initialized before ({@link #getReference()}
	 * returns <tt>null</tt>)
	 * </p>
	 * 
	 * @param port the port to be used by this distributed object when running
	 * as a server
	 * @param daemon tells this server to run as a daemon or a normal server
	 * @throws SdsException raised if the port is already reserved or when this
	 * distributed object has no server ability
	 */
	public abstract void run(int port, boolean daemon) throws SdsException;

	/**
	 * <p>
	 * Executes this distributed object as a normal server (not a daemon). This
	 * operation must be non-blocking. A default reference is created if not
	 * initialized before ({@link #getReference()} returns <tt>null</tt>)
	 * </p>
	 * 
	 * @param port the port to be used by this distributed object when running
	 * as a server
	 * @throws SdsException raised if the port is already reserved or when this
	 * distributed object has no server ability
	 * @see #run(int, boolean)
	 */
	public final void run(int port) throws SdsException {
		run(port, false);
	}

	/**
	 * <p>
	 * Message delivery to recipients
	 * </p>
	 * 
	 * @param msg the message to deliver
	 * @param block <tt>true</tt> if this delivery operation waits for all
	 * messages to be sent and corresponding responses to be received
	 * @param recipients The list of recipients to receive the message
	 * @return the collection of non-delivered recipients. Always empty in case
	 * of non-blocking deliveries. <tt>Null</tt> when the message is
	 * <tt>null</tt> or when no recipient is specified (the provided list of
	 * recipients is either <tt>null</tt> or empty)
	 */
	public final Collection<Reference> deliver(Message msg, boolean block,
		Reference... recipients) {
		if ((msg == null) || (recipients == null) || (recipients.length == 0)) {
			return null;
		}

		Message myMsg = new Message(msg);
		myMsg.setIssuer(getReference());
		MessageData msgData = new MessageData();
		msgData.setMessage(myMsg);

		Long id;
		TransactionInfo info = myMsg.getTransactionInfo();

		// A response from each recipient may be expected
		if (block) {
			if (info == null) {
				info = new TransactionInfo();
				myMsg.setTransactionInfo(info);
			}

			synchronized (expectedResponses) {
				do {
					id = SimpleRandom.nextLong();
				}
				while (expectedResponses.containsKey(id));

				ResponseInfo respInfo = new ResponseInfo();
				respInfo.setCount(new Long(recipients.length));
				respInfo.setSignatureRequired(info.isSignatureRequired());
				expectedResponses.put(id, respInfo);
			}

			info.setNextIdentifier(id);
		}
		else {
			if (info == null) {
				id = null;
			}
			else {
				id = SimpleRandom.nextLong();
				info.setNextIdentifier(id);
			}
		}

		SignedData tbsData = new SignedData();
		tbsData.setTbsData(msgData);
		SecureMessageFormat smf = new SecureMessageFormat();
		smf.setTbsData(tbsData);

		Delivery delivery = new Delivery(smf, recipients);
		delivery.start();

		if (block) {
			try {
				delivery.join();
			}
			catch (InterruptedException e1) {
				return delivery.getInvalidRecipients();
			}

			// Remove non-delivered recipients expected responses
			Collection<Reference> invalidRecipients = delivery
				.getInvalidRecipients();
			ResponseInfo expected = expectedResponses.get(id);
			AtomicLong value = expected.getCount();
			AtomicLong newVal = new AtomicLong(value.longValue()
				- invalidRecipients.size());
			expected.setCount(newVal);
			expectedResponses.put(id, expected);
			expected.lock();
		}

		return delivery.getInvalidRecipients();
	}

	/**
	 * <p>
	 * Message delivery to recipients
	 * </p>
	 * 
	 * @param msg the message to deliver
	 * @param block <tt>true</tt> if this delivery operation waits for all
	 * messages to be sent and corresponding
	 * @param recipients the collection of recipients to receive the message
	 * @return the collection of non-delivered recipients. Always empty in case
	 * of non-blocking deliveries. <tt>Null</tt> when the message is
	 * <tt>null</tt> or the provided collection is <tt>null</tt>
	 */
	public final Collection<Reference> deliver(Message msg, boolean block,
		Collection<Reference> recipients) {
		if (recipients == null) {
			return null;
		}

		Reference[] recipient = new Reference[recipients.size()];
		recipients.toArray(recipient);
		return deliver(msg, block, recipient);
	}

	/**
	 * <p>
	 * Called by the hosting daemon when asked (by a manager) to kill its
	 * running distributed objects. Allows to perform tasks before an effective
	 * closing (along with this distributed object's unregistration from its
	 * local daemon)
	 * </p>
	 */
	public abstract void terminate();
}
