/**
 * Copyright 2007 Nathanael Cottin
 */
package sds;

import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;

import sds.data.Message;
import sds.data.TransactionInfo;
import sds.error.ActivationException;
import sds.error.AuthenticationException;
import sds.error.CommunicationException;
import sds.error.FormatException;
import sds.error.SdsException;
import sds.more.Codec;

/**
 * 
 * @author ncottin
 */
final class DistributedServant extends Thread {

	private final DistributedServerObject	parent;
	private final boolean requiresSignature;

	DistributedServant(DistributedServerObject parent) {
		this.parent = parent;
		requiresSignature = parent.isSignatureRequired();
	}

	/**
	 * <p>
	 * Called by {@link DistributedServer#terminate()} to terminate its
	 * execution
	 * </p>
	 */
	public void terminate() {
		interrupt();
	}
	
	public void terminate(Object lock) {
		interrupt();
	}

	boolean process(Socket client) throws SdsException {
		Object obj;
		try {
			InputStream in = client.getInputStream();
			obj = Codec.decode(in);
			in.close();
			client.close();
		}
		catch (IOException e) {
			throw new CommunicationException(e);
		}
		catch (Exception e) {
			throw new ActivationException(e);
		}

		if (!(obj instanceof SecureMessageFormat)) {
			throw new FormatException();
		}

		SecureMessageFormat smf = (SecureMessageFormat) obj;
		MessageData msgData = getMessageData(smf);
		Message msg = msgData.getMessage();

		// Get transactional identifier
		Long id;
		TransactionInfo info = msg.getTransactionInfo();
		if (info != null) {
			id = info.getLastIdentifier();
			if (id == null) {
				id = info.getNextIdentifier();
			}

			parent.countResponse(id);
		}

		boolean recv = parent.receive(msg);
		return recv;
	}

	// Incoming message authentication procedure
	private MessageData getMessageData(SecureMessageFormat smf) throws AuthenticationException, FormatException {
		SignedData signedData = smf.getTbsData();
		if (signedData == null) {
			throw new FormatException("No tbsData");
		}
		
		MessageData msgData = signedData.getTbsData();
		if (msgData == null) {
			throw new FormatException("No message data");
		}
		
		byte[] signValue = smf.getSignatureValue();
		if ((signValue == null) || (signValue.length == 0)) {
			if (requiresSignature) {
				throw new AuthenticationException("No signature value");
			}
		}
		else {
			// Validate signature over SignedData value
			// TODO
		}
		
		if (!parent.getReference().equals(msgData.getRecipient())) {
			throw new AuthenticationException("Invalid recipient reference");
		}

		return msgData;
	}
}
