/**
 * Copyright 2007 Nathanael Cottin
 */
package sds;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import sds.error.CommunicationException;
import sds.error.ProcessException;
import sds.error.SdsException;

/**
 * @author ncottin
 */
final class DistributedServer extends Thread {

	private static final Boolean			FALSE	= new Boolean(false);
	private static final Boolean			TRUE	= new Boolean(true);
	private static final long				DELAY	= 100;
	private final DistributedServerObject	parent;
	private ServerSocket					sock;
	private final Queue<DistributedServant>	queue;
	private AtomicInteger					started;
	private Boolean							alive;
	private final Object					lock;

	public DistributedServer(DistributedServerObject parent, int port)
		throws SdsException {
		try {
			sock = new ServerSocket(port);
		}
		catch (IOException e) {
			throw new CommunicationException(e);
		}

		this.parent = parent;
		queue = new ConcurrentLinkedQueue<DistributedServant>();
		started = new AtomicInteger(0);
		alive = TRUE;
		lock = new Object();
	}

	@Override
	public void run() {
		Socket client = null;
		DistributedServant servant;
		boolean available;
		while (alive.booleanValue()) {
			try {
				client = sock.accept();
				available = true;
			}
			catch (SocketException e) {
				synchronized (alive) {
					alive = FALSE;
				}

				available = false;
			}
			catch (IOException e) {
				available = false;
				parent.processError(new CommunicationException(e));
			}

			if (available) {
				servant = getServant(parent);
				try {
					if (!servant.process(client)) {
						throw new ProcessException();
					}
				}
				catch (SdsException e) {
					parent.processError(e);
				}

				endServant(servant);
			}
		}
	}

	private DistributedServant getServant(DistributedServerObject parent) {
		started.incrementAndGet();
		if (queue.isEmpty()) {
			DistributedServant servant = new DistributedServant(parent);
			servant.start();
			return servant;
		}

		return queue.poll();
	}

	private void endServant(DistributedServant idle) {
		queue.add(idle);
		if (started.decrementAndGet() == 0) {
			synchronized (lock) {
				lock.notify();
			}
		}
	}

	public void terminate() {
		synchronized (alive) {
			alive = FALSE;
		}

		// Wait for all active servants to terminate
		while (started.intValue() > 0) {
			synchronized (lock) {
				try {
					lock.wait();
				}
				catch (InterruptedException e) {
					// Forget it
				}
			}
		}

		try {
			sock.close();
		}
		catch (IOException e) {
		}

		// It is necessary to put a loop as servant threads would add themselves
		// in the queue after the last (which may not be actually!)
		// 'queue.poll()' call
		DistributedServant servant;
		do {
			while ((servant = queue.poll()) != null) {
				servant.terminate();
			}

			try {
				Thread.sleep(DELAY);
			}
			catch (InterruptedException e) {
				break;
			}
		}
		while (!queue.isEmpty());
	}
}
