/**
 * Copyright 2007 Nathanael Cottin
 */
package sds;

import sds.data.Message;
import sds.data.Reference;
import sds.error.SdsException;
import sds.more.Network;

/**
 * <p>
 * Provides server-oriented distributed objects functionalities. A
 * server-oriented distributed object waits for incoming messages on its
 * allocated port number
 * </p>
 * 
 * @author ncottin
 */
public abstract class DistributedServerObject extends DistributedObject {

	private final boolean		requiresSignature;
	private DistributedServer	server	= null;

	/**
	 * <p>
	 * Creates a server-oriented distributed object. A default reference is
	 * automatically created. This distributed object is not associated with a
	 * manager
	 * </p>
	 * 
	 * @param requiresSignature <tt>true</tt> if all incoming non digitally
	 * signed (and authenticated) messages must be rejected
	 */
	public DistributedServerObject(boolean requiresSignature) {
		this.requiresSignature = requiresSignature;
	}

	public final boolean isSignatureRequired() {
		return requiresSignature;
	}

	/**
	 * <p>
	 * Checks if this distributed object can receive message
	 * </p>
	 * 
	 * @return <tt>true</tt> as this distributed object is designed to behave
	 * as a server
	 */
	@Override
	public final boolean runsAsServer() {
		return true;
	}

	@Override
	public final void run(int port, boolean daemon) throws SdsException {
		server = new DistributedServer(this, port);

		// Complete server-oriented distributed object's reference with IP
		// address and port number
		Reference reference = getReference();
		if (reference == null) {
			setDefaultReference();
			reference = getReference();
		}

		reference.setAddress(Network.getIpAddress());
		reference.setPort(port);
		server.setDaemon(daemon);
		server.start();
		execute();
	}

	/**
	 * <p>
	 * Defines this server's default behaviour (i.e. while not receiving
	 * messages from other distributed objects)
	 * </p>
	 */
	public abstract void execute();

	/**
	 * <p>
	 * Called by the hosting daemon when asked (by a manager) to kill its
	 * running distributed objects. Allows to perform tasks before an effective
	 * closing (along with freeing this distributed object). This operation must
	 * not be called within a thread (mainly from {@link #receive(Message)}) as
	 * it would lead to wait indefinitely (until the current thread stops, which
	 * never occurs!)
	 * </p>
	 */
	@Override
	public final void terminate() {
		terminateServer();
		// mainExec.terminate();
		server.terminate();
	}

	/**
	 * <p>
	 * Automatically called by {@link #terminate()}
	 * </p>
	 */
	public abstract void terminateServer();

	/**
	 * <p>
	 * Gets the server thread
	 * </p>
	 * 
	 * @return the internal main server thread. Set to <tt>null</tt> until
	 * {@link #run(int)} is called
	 */
	DistributedServer getServer() {
		return server;
	}

	/**
	 * <p>
	 * Defines the server behavior when a message is received
	 * </p>
	 * 
	 * @param msg received message. Must not be <tt>null</tt> (no checking is
	 * necessary when implementing this operation)
	 * @throws SdsException an error occurred when retrieving information from
	 * the message (or cast)
	 * @return <tt>true</tt> if request message content is successfully
	 * handled
	 */
	public abstract boolean receive(Message msg) throws SdsException;

	/**
	 * <p>
	 * Automatically called when {@link #receive(Message)} returns
	 * <tt>false</tt> or when an {@link sds.error.SdsException} while
	 * processing an incoming request
	 * </p>
	 * 
	 * @param error
	 */
	public abstract void processError(SdsException error);
}
