/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.data;

import java.io.Serializable;

/**
 * 
 */

/**
 * <p>
 * Information sent to and received by distant servers. This information is to
 * be encapsulated within a top-level envelope
 * </p>
 * 
 * @author ncottin
 */
public final class Message implements Serializable {

	private Reference		issuer			= null;
	private TransactionInfo	transactionInfo	= null;
	private Integer			tag				= null;
	private Object			data			= null;

	/**
	 * <p>
	 * Creates an empty message. This constructor should not be used to create a
	 * response to an exiting (request) message
	 * </p>
	 * 
	 * @see #createResponse()
	 */
	public Message() {
	}

	/**
	 * <p>
	 * Duplicates an existing message. As all fields but data are duplicated,
	 * this constructor should not be used to create a response to an exiting
	 * (request) message
	 * </p>
	 * 
	 * @param other The source message to duplicate. The current message is left
	 * empty in case the other message is <tt>null</tt>
	 * @see #createResponse()
	 */
	public Message(Message other) {
		if (other == null) {
			return;
		}

		issuer = new Reference(other.issuer);
		transactionInfo = new TransactionInfo(other.transactionInfo);
		if (other.tag != null) {
			tag = new Integer(other.tag.intValue());
		}

		data = other.data;
	}

	/**
	 * <p>
	 * Generates a pre-initialized response to this message. This newly created
	 * message basically encloses default (duplicated) transactional information
	 * </p>
	 * 
	 * @return A non-<tt>null</tt> and uninitialized message but duplicated
	 * transactional information (if set in the current message)
	 * @see TransactionInfo#createResponse()
	 */
	public Message createResponse() {
		Message resp = new Message();
		TransactionInfo info = (transactionInfo == null)
			? null
			: transactionInfo.createResponse();
		resp.setTransactionInfo(info);
		return resp;
	}

	/**
	 * <p>
	 * Generates a pre-initialized response to a given message. The newly
	 * created message basically encloses default (duplicated) transactional
	 * information
	 * </p>
	 * 
	 * @param request The request message to respond to
	 * @return A non-<tt>null</tt> and uninitialized message but duplicated
	 * transactional information (if set in the current message)
	 * @see #createResponse()
	 */
	public static Message createResponse(Message request) {
		return (request == null)
			? null
			: request.createResponse();
	}

	/**
	 * <p>
	 * Gets the message issuer
	 * </p>
	 * 
	 * @return The message issuer (sender) reference. This reference may be
	 * <tt>null</tt> if the message is not initialized
	 */
	public Reference getIssuer() {
		return issuer;
	}

	/**
	 * <p>
	 * Assigns the message issuer reference
	 * </p>
	 * 
	 * @param issuer The issuer reference. The message cannot be sent until this
	 * reference is provided
	 */
	public void setIssuer(Reference issuer) {
		this.issuer = issuer;
	}

	/**
	 * <p>
	 * Gets the information related with transactions (requests / responses
	 * track identifiers)
	 * </p>
	 * 
	 * @return The transactional information or <tt>null</tt> if not such
	 * information is provided
	 */
	public TransactionInfo getTransactionInfo() {
		return transactionInfo;
	}

	/**
	 * <p>
	 * Updates the transactional information of this message (mainly based on a
	 * previous received message in case of a response)
	 * </p>
	 * 
	 * @param transactionInfo The transactional information or <tt>null</tt>
	 * if no transaction information applies to this message
	 * @see #createResponse()
	 */
	public void setTransactionInfo(TransactionInfo transactionInfo) {
		this.transactionInfo = transactionInfo;
	}

	/**
	 * <p>
	 * Gets the current message tag
	 * </p>
	 * 
	 * @return Any value, including <tt>null</tt> if this message has no
	 * assigned tag
	 */
	public Integer getTag() {
		return tag;
	}

	/**
	 * <p>
	 * Initializes the tag of this message
	 * </p>
	 * 
	 * @param tag Any value. A <tt>null</tt> tag leads
	 * {@link #isInitialized()} to return <tt>false</tt>
	 */
	public void setTag(Integer tag) {
		this.tag = tag;
	}

	/**
	 * <p>
	 * Gets the message content
	 * </p>
	 * 
	 * @return The message main content
	 */
	public Object getData() {
		return data;
	}

	/**
	 * <p>
	 * Initializes the message content
	 * </p>
	 * 
	 * @param data The message content
	 */
	public void setData(Object data) {
		this.data = data;
	}

	/**
	 * <p>
	 * Makes sure this message can be sent
	 * </p>
	 * 
	 * @return <tt>True</tt> only if this message encloses a tag value and the
	 * reference to the sender
	 */
	public boolean isInitialized() {
		return (tag != null) && (issuer != null);
	}

	/**
	 * <p>
	 * Checks whether this message encloses transactional information
	 * </p>
	 * 
	 * @return <tt>True</tt> only if this message transactional information is
	 * not <tt>null</tt> and initialized
	 */
	public boolean isTransactional() {
		return (transactionInfo != null) && transactionInfo.isInitialized();
	}
}
