/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.data;

import java.io.Serializable;
import java.util.StringTokenizer;

/**
 * <p>
 * Mean for identifying distributed objects and providing messages delivery
 * contact information
 * </p>
 * 
 * @author ncottin
 */
public final class Reference implements Serializable {

	public static final long	DEFAULT_VALUE	= 0;
	public static final int		NO_PORT			= 0;
	private static final String	NULL_ADDRESS	= "-";
	private static final String	SEP				= ":";
	private static final int	NB_ELEMENTS		= 4;
	private long				id				= DEFAULT_VALUE;
	private long				nonce			= DEFAULT_VALUE;
	private String				address			= null;
	private int					port			= NO_PORT;

	/**
	 * <p>
	 * Creates an uninitialized reference
	 * </p>
	 */
	public Reference() {
	}

	/**
	 * <p>
	 * Creates and initializes this reference with another reference
	 * </p>
	 * 
	 * @param other A <tt>null</tt> reference leaves this reference
	 * uninitialized. This reference duplicates the given reference otherwise
	 */
	public Reference(Reference other) {
		if (other == null) {
			return;
		}

		id = other.id;
		nonce = other.nonce;
		address = other.address;
		port = other.port;
	}

	/**
	 * <p>
	 * Initializes from a string representation
	 * </p>
	 * 
	 * @param value
	 * @return <tt>True</tt> only if successfully decoded from string.
	 * <tt>False</tt> implies that the current reference values are left
	 * unchanged
	 */
	public boolean fromString(String value) {
		if ((value == null) || "".equals(value)) {
			return false;
		}

		StringTokenizer tok = new StringTokenizer(value, SEP);
		if (tok.countTokens() != NB_ELEMENTS) {
			return false;
		}

		long id;
		long nonce;
		String address;
		int port;
		try {
			id = Long.parseLong(tok.nextToken());
			nonce = Long.parseLong(tok.nextToken());
			address = tok.nextToken();
			port = Integer.parseInt(tok.nextToken());
		}
		catch (NumberFormatException e) {
			return false;
		}

		this.id = id;
		this.nonce = nonce;
		this.address = address;
		this.port = port;
		return true;
	}

	/**
	 * <p>
	 * Gets this reference's identifier
	 * </p>
	 * 
	 * @return Any value
	 */
	public long getId() {
		return id;
	}

	/**
	 * <p>
	 * Assigns an identifier to this reference
	 * </p>
	 * 
	 * @param id A positive id
	 */
	public void setId(long id) {
		this.id = id;
	}

	/**
	 * @return The nonce randomly assigned to this reference
	 */
	public long getNonce() {
		return nonce;
	}

	/**
	 * @param nonce Any random value
	 */
	public void setNonce(long nonce) {
		this.nonce = nonce;
	}

	/**
	 * @return the address
	 */
	public String getAddress() {
		return address;
	}

	/**
	 * <p>
	 * Initializes the IP address with a given address
	 * </p>
	 * 
	 * @param address The host IP address. Should neither be <tt>null</tt> nor
	 * empty to initialize this reference with
	 */
	public void setAddress(String address) {
		this.address = address;
	}

	/**
	 * <p>
	 * The port assigned to the referenced distributed object. This port is
	 * valid only when referencing a running server-oriented object
	 * </p>
	 * 
	 * @return A strictly positive port refers to a running server-oriented
	 * distributed object. A negative value (such as {@link #NO_PORT})
	 * indicates either a client-oriented or a non-running server-oriented
	 * distributed object
	 */
	public int getPort() {
		return port;
	}

	/**
	 * @param port The port number. Must be strictly positive to be valid
	 */
	public void setPort(int port) {
		this.port = port;
	}

	/**
	 * <p>
	 * Checks whether this reference points to a client-oriented or a
	 * server-oriented distributed object
	 * </p>
	 * 
	 * @return <tt>True</tt> only of this reference belongs to a
	 * server-oriented distributed object (i.e. contains a non-<tt>null</tt>
	 * and non-empty IP address and a valid port number)
	 */
	public boolean isInvokable() {
		return (port > NO_PORT) && (address != null) && !"".equals(address);
	}

	@Override
	public boolean equals(Object other) {
		if ((other == null) || !(other instanceof Reference)) {
			return false;
		}

		Reference ref = (Reference) other;
		return (id == ref.id)
			&& (nonce == ref.nonce)
			&& (port == ref.port)
			&& (((address == null) && (ref.address == null)) || address
				.equals(ref.address));
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(address);
		sb.append(SEP);
		sb.append(port);
		sb.append(SEP);
		sb.append(id);
		sb.append(SEP);
		sb.append(nonce);

		// StringBuffer append the string value "null" when a null string is
		// given
		// => The hashing process must make a difference between null and "null"
		// addresses
		if (address == null) {
			sb.append(NULL_ADDRESS);
		}

		return sb.toString();
	}

	@Override
	public int hashCode() {
		return toString().hashCode();
	}
}
