/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.data;

import java.io.Serializable;

/**
 * <p>
 * Message transaction information in terms of requests / responses correlation,
 * expressed by the combination of <tt>last</tt> and <tt>next</tt>
 * identifiers.
 * </p>
 * <p>
 * A typical transaction between two server-oriented distributed objects <i>A</i>
 * and <i>B</i> is described as follows:
 * </p>
 * <ol>
 * <li><i>A</i> creates a request message including transactional information.
 * It randomly generates an identifier to initialize <tt>next</tt> identifier
 * with</li>
 * <li><i>A</i> sends the transactional message to <i>B</i></li>
 * <li><i>B</i> receives the message from <i>A</i>. It checks that this
 * message contains transactional information. If so, <i>B</i> should produce a
 * response</li>
 * <li><i>B</i> creates a response message and assigns a transactional
 * information in return. This transactional information duplicates the received
 * <tt>next</tt> identifier as its <tt>last</tt> field value</li>
 * <li><i>A</i> receives a message from <i>B</i>. It checks if it contains
 * transactional information to determine to which request this message responds
 * to</li>
 * </ol>
 * <p>
 * Finally, <i>B</i> could also have set a <tt>next</tt> identifier value if
 * a response from <i>A</i> is needed...
 * </p>
 * 
 * @author ncottin
 */
public final class TransactionInfo implements Serializable {

	private Long	lastIdentifier	= null;
	private Long	nextIdentifier	= null;
	private boolean	requiresSignature;

	/**
	 * <p>
	 * Creates an empty transaction information. {@link #isInitialized()}
	 * returns <tt>false</tt>
	 * </p>
	 */
	public TransactionInfo() {
		requiresSignature = false;
	}

	/**
	 * <p>
	 * Creates and initializes this transaction information with another
	 * information
	 * </p>
	 * 
	 * @param other The existing transaction information to be duplicated.
	 * <tt>Null</tt> indicates that this transaction information is left
	 * uninitialized
	 */
	public TransactionInfo(TransactionInfo other) {
		if (other == null) {
			return;
		}

		Long val = other.getLastIdentifier();
		if (val != null) {
			lastIdentifier = new Long(val.longValue());
		}

		val = other.getNextIdentifier();
		if (val != null) {
			nextIdentifier = new Long(val.longValue());
		}

		requiresSignature = other.requiresSignature;
	}

	/**
	 * <p>
	 * Produces a pre-filled transactional information element from this
	 * transactional information (which should be taken from a request message)
	 * </p>
	 * 
	 * @return A newly created transactional information object containing a
	 * copy of the current <tt>next</tt> identifier assigned as <tt>last</tt>
	 * identifier. This transactional information is left empty (uninitialized)
	 * in case the current transactional information does not hold any
	 * <tt>next</tt> identifier value
	 */
	public TransactionInfo createResponse() {
		TransactionInfo res = new TransactionInfo();
		if (nextIdentifier != null) {
			res.setLastIdentifier(nextIdentifier);
		}

		return res;
	}

	/**
	 * <p>
	 * Produces a pre-filled transactional information element from a given
	 * transactional information (which should be taken from a request message)
	 * </p>
	 * 
	 * @param source The source transaction information (usually taken from a
	 * request message) to be used to create the resulting transactional
	 * information
	 * @return A newly created transactional information object containing a
	 * copy of the provided transactional information's <tt>next</tt>
	 * identifier assigned as <tt>last</tt> identifier. This transactional
	 * information is left empty (uninitialized) if either the given
	 * transactional information is <tt>null</tt> or does not hold any
	 * <tt>next</tt> identifier value
	 * @see #createResponse()
	 */
	public static TransactionInfo createResponse(TransactionInfo source) {
		return (source == null)
			? null
			: source.createResponse();
	}

	/**
	 * <p>
	 * Gets the <tt>last</tt> identifier
	 * </p>
	 * 
	 * @return The last known identifier or <tt>null</tt> if this transaction
	 * does not hold any <tt>last</tt> identifier
	 */
	public Long getLastIdentifier() {
		return lastIdentifier;
	}

	/**
	 * <p>
	 * Assigns the <tt>last</tt> identifier value in case this transactional
	 * information takes part of a response of a transactional request message.
	 * This value must be taken from the latter and corresponds to its
	 * <tt>next</tt> identifier
	 * </p>
	 * 
	 * @param lastIdentifier The last transactional identifier or <tt>null</tt>
	 * for no identifier
	 * @see #createResponse()
	 */
	public void setLastIdentifier(Long lastIdentifier) {
		this.lastIdentifier = lastIdentifier;
	}

	/**
	 * <p>
	 * Gets the <tt>next</tt> identifier
	 * </p>
	 * 
	 * @return The next identifier to be returned back by the recipient of the
	 * message (which includes this transactional information) or <tt>null</tt>
	 * if this transaction does not hold any <tt>next</tt> identifier
	 */
	public Long getNextIdentifier() {
		return nextIdentifier;
	}

	/**
	 * <p>
	 * Assigns the <tt>next</tt> identifier so that the recipient can assign
	 * its <tt>last</tt> identifier value when generating a reply. This value
	 * should be randomly generated
	 * </p>
	 * 
	 * @param nextIdentifier The next transactional identifier or <tt>null</tt>
	 * for no identifier. This identifier is returned back by the enclosing
	 * message recipients
	 */
	public void setNextIdentifier(Long nextIdentifier) {
		this.nextIdentifier = nextIdentifier;
	}

	/**
	 * <p>
	 * Checks that this transactional information contains a <tt>next</tt>
	 * identifier
	 * </p>
	 * 
	 * @return <tt>True</tt> only if the <tt>next</tt> identifier is not
	 * <tt>null</tt>
	 */
	public boolean hasNextIdentifier() {
		return nextIdentifier != null;
	}

	/**
	 * <p>
	 * Checks that this transactional information contains a <tt>last</tt>
	 * identifier
	 * </p>
	 * 
	 * @return <tt>True</tt> only if the <tt>last</tt> identifier is not
	 * <tt>null</tt>
	 */
	public boolean hasLastIdentifier() {
		return lastIdentifier != null;
	}

	/**
	 * @return <tt>true</tt> means that this transactional information
	 * requires the answer to be digitally signed
	 */
	public boolean isSignatureRequired() {
		return requiresSignature;
	}

	/**
	 * @param requiresSignature <tt>true</tt> only if this transactional
	 * information requires the answer to be digitally signed
	 */
	public void setSignatureRequired(boolean requiresSignature) {
		this.requiresSignature = requiresSignature;
	}

	/**
	 * <p>
	 * Makes sure this transactional information is initialized before inclusion
	 * in a message
	 * </p>
	 * 
	 * @return <tt>True</tt> only if this transactional information holds at
	 * least one non-<tt>null</tt> identifier
	 */
	public boolean isInitialized() {
		return (lastIdentifier != null) || (nextIdentifier != null);
	}
}
