/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.more;

import java.util.Collection;
import java.util.HashSet;

import sds.data.Reference;

/**
 * <p>
 * Unclassifiable helpful operations either used by SDS or for the developer
 * </p>
 * 
 * @author ncottin
 */
public final class Miscellaneous {

	/**
	 * <p>
	 * Transforms a collection of references to an array of references,
	 * discarding all <tt>null</tt> references
	 * </p>
	 * 
	 * @param references A possibly <tt>null</tt> collection of references
	 * @return <tt>Null</tt> if the given collection is <tt>null</tt> or a
	 * possibly empty collection of references holding all non-<tt>null</tt>
	 * references taken from the provided collection
	 */
	public static Reference[] toArray(Collection<Reference> references) {
		if (references == null) {
			return null;
		}

		Collection<Reference> refs = new HashSet<Reference>(references.size());
		for (Reference ref : references) {
			if (ref != null) {
				refs.add(ref);
			}
		}

		Reference[] res = new Reference[refs.size()];
		refs.toArray(res);
		return res;
	}

	/**
	 * <p>
	 * Transforms a list of references to a collection of references, discarding
	 * <tt>null</tt> references
	 * </p>
	 * 
	 * @param references A possibly <tt>null</tt> list of references
	 * @return <tt>Null</tt> if the given list is <tt>null</tt> or a
	 * possibly empty collection of references holding all non-<tt>null</tt>
	 * references
	 */
	public static Collection<Reference> toCollection(Reference... references) {
		if (references == null) {
			return null;
		}

		Collection<Reference> res = new HashSet<Reference>(references.length);
		for (Reference ref : references) {
			if (ref != null) {
				res.add(ref);
			}
		}

		return res;
	}
}
