/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.service;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;

import sds.DistributedObject;
import sds.data.Message;
import sds.data.Reference;
import sds.error.ActivationException;
import sds.error.CommunicationException;
import sds.error.SdsException;
import sds.more.SimpleClassLoader;

/**
 * <p>
 * Server-oriented distributed object to be started on each distant host which
 * participates to the distributed system. It receives distributed objects sent
 * by a {@link Manager} and manages their life cycle (from load and activation
 * to destruction)
 * </p>
 * 
 * @see sds.DistributedObject
 * @author ncottin
 */
public final class Daemon extends CommonService {

	/**
	 * Default identifier (= <tt>0</tt>) for loaded distributed objects
	 */
	public static final long					DEFAULT_FIRST_ID	= 0;

	/**
	 * Indicates that this daemon can load and run server-oriented distributed
	 * objects without taking care of available ports numbers
	 */
	public static final int						NO_MAX_PORT			= -1;
	private final SimpleClassLoader				loader;
	private final int							minPort;
	private final int							maxPort;
	private final boolean						maxPortEnabled;
	private Integer								nextPort;
	private Long								nextId;
	private final Map<Long, DistributedObject>	regDistObjects;

	public Daemon(boolean requiresSignature, int minPort, int maxPort,
		long firstId) {
		super(requiresSignature);
		loader = new SimpleClassLoader();
		this.minPort = minPort;
		this.maxPort = maxPort;
		maxPortEnabled = maxPort >= minPort;
		nextPort = new Integer(minPort);
		nextId = new Long(firstId);
		regDistObjects = new Hashtable<Long, DistributedObject>();
	}

	public int getMinPort() {
		return minPort;
	}

	public int getMaxPort() {
		return maxPort;
	}

	public boolean hasMaxPort() {
		return maxPortEnabled;
	}

	@Override
	public void execute() {
		// Nothing to execute
	}

	@Override
	public void processError(SdsException e) {
		System.out.println("Daemon encountered the following error:");
		e.printStackTrace();
	}

	@Override
	public boolean receive(Message msg) throws SdsException {
		Message resp;
		int tag = msg.getTag();
		switch (tag) {
			case CREATE_TAG_REQUEST:
				resp = createObject(msg);
				break;
			case KILL_TAG_REQUEST:
				resp = removeObjects(msg);
				break;
			default:
				return false;
		}

		if (resp == null) {
			return false;
		}

		// 'block' is set to false as no further response is expected
		return deliver(resp, false, msg.getIssuer()).isEmpty();
	}

	@SuppressWarnings("unchecked")
	private Message createObject(Message msg) {
		Message resp = msg.createResponse();
		resp.setTag(CREATE_TAG_RESPONSE);

		Object obj = msg.getData();
		if ((obj == null) || !(obj instanceof ClassData)) {
			return resp;
		}

		ClassData classData = (ClassData) obj;
		obj = null;
		int nbInstances = classData.getNbInstances();
		if (nbInstances < 1) {
			return resp;
		}

		Reference mgr = msg.getIssuer();
		Collection<Reference> refs = new HashSet<Reference>(nbInstances);
		Class<? extends DistributedObject> cl = null;
		try {
			cl = (Class<? extends DistributedObject>) loader.loadClass(
				classData.getName(), classData.getByteCode(), true);
			DistributedObject distObj;
			boolean reserved;
			Reference ref;
			long id;
			for (int i = 0; i < nbInstances; i++) {
				distObj = instantiate(cl);
				distObj.setManager(mgr);
				if (distObj.runsAsServer()) {
					// First, try to use next port
					if (maxPortEnabled && (nextPort > maxPort)) {
						throw new ActivationException("No port available");
					}

					try {
						synchronized (nextPort) {
							distObj.run(nextPort.intValue(), false);
							nextPort++;
							reserved = false;
						}

						if (false) {
							throw new CommunicationException();
						}
					}
					catch (SdsException exc) {
						reserved = true;
					}

					// If next port is not available,
					// then try again from minimum port
					// until an available port is found
					if (reserved) {
						synchronized (nextPort) {
							nextPort = minPort;
							do {
								try {
									distObj.run(nextPort.intValue(), false);
									reserved = false;
								}
								catch (SdsException exc) {
									nextPort++;
									if (maxPortEnabled && (nextPort > maxPort)) {
										throw new ActivationException(
											"No port available");
									}
								}
							}
							while (reserved);
						}
					}

					synchronized (nextId) {
						id = nextId;
						nextId++;
					}

					ref = distObj.getReference();
					ref.setId(id);
					refs.add(ref);
					regDistObjects.put(id, distObj);
				}
			}
		}
		catch (Exception e) {
		}

		if (cl != null) {
			ReferencedObject refObj = new ReferencedObject();
			refObj.setId(cl);
			refObj.setReferences(refs);
			resp.setData(refObj);
		}

		return resp;
	}

	private DistributedObject instantiate(Class<? extends DistributedObject> cl)
		throws InstantiationException,
			IllegalAccessException {
		DistributedObject distObj = (DistributedObject) cl.newInstance();
		return distObj;
	}

	@SuppressWarnings("unchecked")
	private Message removeObjects(Message msg) {
		Message resp = msg.createResponse();
		resp.setTag(KILL_TAG_RESPONSE);

		Object obj = msg.getData();
		if (obj == null) {
			return resp;
		}

		Collection<Reference> refs;
		try {
			refs = (Collection<Reference>) msg.getData();
		}
		catch (ClassCastException e) {
			return resp;
		}

		if (refs == null) {
			return resp;
		}

		obj = null;
		if (refs.isEmpty()) {
			resp.setData(refs);
			return resp;
		}

		Collection<Reference> deletedRefs = new ArrayList(refs.size());
		Long id;
		for (Reference ref : refs) {
			if (ref != null) {
				id = ref.getId();
				DistributedObject distObj = regDistObjects.get(id);
				if (distObj != null) {
					distObj.terminate();
					Reference regRef = distObj.getReference();
					if (ref.equals(regRef)
						&& (regDistObjects.remove(id) != null)) {
						deletedRefs.add(regRef);
					}
				}
			}
		}

		resp.setData(deletedRefs);
		return resp;
	}

	@Override
	public void terminateServer() {

	}
}