/**
 * Copyright 2007 Nathanael Cottin
 */
package sds.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;

import sds.DistributedObject;
import sds.data.Message;
import sds.data.Reference;
import sds.error.ActivationException;
import sds.error.CommunicationException;
import sds.error.SdsException;
import sds.more.SimpleClassLoader;

/**
 * <p>
 * Server-oriented distributed object to be instantiated by the user-defined
 * distributed system manager. Registers a set of participating {@link Daemon}
 * distributed objects, sends user-defined distributed objects' bytecodes and
 * keeps the latter's references
 * </p>
 * 
 * @see sds.DistributedObject
 * @author ncottin
 */
public abstract class Manager extends CommonService {

	private final Collection<Reference>																registeredDaemons;
	private final Map<Class<? extends DistributedObject>, Map<Reference, Collection<Reference>>>	objReferences;

	/**
	 * <p>
	 * Creates a manager with no registered daemon nor distributed objects
	 * references. This manager holds no associated reference
	 * </p>
	 * 
	 * @param requiresSignature <tt>true</tt> if all incoming non digitally
	 * signed messages must be rejected
	 */
	public Manager(boolean requiresSignature) {
		super(requiresSignature);
		registeredDaemons = Collections
			.synchronizedSet(new HashSet<Reference>());
		objReferences = new Hashtable<Class<? extends DistributedObject>, Map<Reference, Collection<Reference>>>();
	}

	/**
	 * <p>
	 * Retrieves all daemons references from a URL. This URL must point to a
	 * list of other URLs leading to daemons references. These references must
	 * be created by means of {@link sds.data.Reference#toString()}
	 * </p>
	 * 
	 * @param locations The URLs leading to the list of daemons' references URLs
	 * @return <tt>True</tt> if at least one invokable reference could be
	 * obtained from the given URL. <tt>False</tt> is returned either if the
	 * provided URL list is not valid (<tt>null</tt> or empty) or if all
	 * referenced URLs are invalid (including <tt>null</tt> values)
	 */
	public final boolean registerDaemons(URL... locations) {
		if ((locations == null) || (locations.length == 0)) {
			return false;
		}

		URLConnection conn;
		BufferedReader br;
		Collection<URL> urls;
		Reference ref;
		Collection<Reference> refs = new HashSet<Reference>();
		for (URL location : locations) {
			// Step 1: read list of daemons references urls
			try {
				conn = location.openConnection();
				br = new BufferedReader(new InputStreamReader(conn
					.getInputStream()));
				urls = new HashSet<URL>();
				String line;
				while ((line = br.readLine()) != null) {
					urls.add(new URL(line));
				}

				br.close();
			}
			catch (IOException e) {
				return false;
			}

			// Step 2: add corresponding references
			for (URL url : urls) {
				try {
					conn = url.openConnection();
					br = new BufferedReader(new InputStreamReader(conn
						.getInputStream()));
					ref = new Reference();
					if (ref.fromString(br.readLine())) {
						refs.add(ref);
					}
				}
				catch (IOException e) {
				}
			}
		}

		return registerDaemons(refs);
	}

	public final boolean registerDaemons(Reference... daemons) {
		if ((daemons == null) || (daemons.length == 0)) {
			return false;
		}

		boolean added = false;
		for (Reference ref : daemons) {
			if ((ref != null) && ref.isInvokable()) {
				registeredDaemons.add(ref);
				added = true;
			}
		}

		return added;
	}

	public final boolean registerDaemons(Collection<Reference> daemons) {
		if (daemons == null) {
			return false;
		}

		Reference[] dms = new Reference[daemons.size()];
		daemons.toArray(dms);
		return registerDaemons(dms);
	}

	/**
	 * <p>
	 * Gets the registered daemons references
	 * </p>
	 * 
	 * @return A non-<tt>null</tt> (but possibly empty) collection of
	 * invokable references (that should correspond to daemons)
	 * @see #registerDaemons(URL...)
	 * @see #registerDaemons(Reference...)
	 */
	public final Collection<Reference> getRegisteredDaemons() {
		return registeredDaemons;
	}

	/**
	 * <p>
	 * Removes all registered daemons references
	 * </p>
	 */
	public final void removeRegisteredDaemons() {
		registeredDaemons.clear();
	}

	/**
	 * <p>
	 * Removes the given list of daemons references from the registered
	 * references
	 * </p>
	 * 
	 * @param daemons Any list of references
	 * @return <tt>True</tt> only if at least one reference was removed
	 */
	public final boolean removeRegisteredDaemons(Reference... daemons) {
		if ((daemons == null) || (daemons.length == 0)) {
			return false;
		}

		boolean removed = false;
		for (Reference ref : daemons) {
			if (ref != null) {
				if (registeredDaemons.remove(ref)) {
					removed = true;
				}
			}
		}

		return removed;
	}

	/**
	 * <p>
	 * Removes the given collection of daemons references from the registered
	 * references
	 * </p>
	 * 
	 * @param daemons Any collection of references
	 * @return <tt>True</tt> only if at least one reference was removed
	 */
	public final boolean removeRegisteredDaemons(Collection<Reference> daemons) {
		if (daemons == null) {
			return false;
		}

		Reference[] dms = new Reference[daemons.size()];
		daemons.toArray(dms);
		return removeRegisteredDaemons(dms);
	}

	@Override
	public final boolean receive(Message msg) throws SdsException {
		Reference issuer = msg.getIssuer();
		if (registeredDaemons.contains(issuer)) {
			int tag = msg.getTag();
			switch (tag) {
				case CREATE_TAG_RESPONSE: {
					return addReferences(msg);
				}
				case KILL_TAG_RESPONSE: {
					return removeReferences(msg);
				}
				default:
					return false;
			}
		}

		return recv(msg);
	}

	public abstract boolean recv(Message msg) throws SdsException;

	private boolean addReferences(Message msg) {
		Object obj = msg.getData();
		if ((obj == null) || !(obj instanceof ReferencedObject)) {
			return false;
		}

		ReferencedObject objRef = (ReferencedObject) obj;
		if ((objRef == null) || !objRef.isInitialized()) {
			return false;
		}

		obj = null;
		Class<? extends DistributedObject> objClass = objRef.getId();
		Map<Reference, Collection<Reference>> regRefs = objReferences
			.get(objClass);
		if (regRefs == null) {
			regRefs = new Hashtable<Reference, Collection<Reference>>();
			objReferences.put(objClass, regRefs);
		}

		Reference issuer = msg.getIssuer();
		Collection<Reference> refs = regRefs.get(issuer);
		if (refs == null) {
			refs = Collections.synchronizedSet(new HashSet<Reference>());
			regRefs.put(issuer, refs);
		}

		return refs.addAll(objRef.getReferences());
	}

	@SuppressWarnings("unchecked")
	private boolean removeReferences(Message msg) {
		Collection<Reference> refs;
		try {
			refs = (Collection<Reference>) msg.getData();
		}
		catch (ClassCastException e) {
			return false;
		}

		if (refs == null) {
			return false;
		}

		if (refs.isEmpty()) {
			return true;
		}

		Reference issuer = msg.getIssuer();
		Map<Reference, Collection<Reference>> regRefs;
		Collection<Reference> issuerRefs;
		for (Class<? extends DistributedObject> cl : objReferences.keySet()) {
			regRefs = objReferences.get(cl);
			if (regRefs != null) {
				issuerRefs = regRefs.get(issuer);
				if (issuerRefs != null) {
					issuerRefs.removeAll(refs);
					if (issuerRefs.isEmpty()) {
						regRefs.remove(issuer);
						if (regRefs.isEmpty()) {
							objReferences.remove(cl);
						}
					}
				}
			}
		}

		return true;
	}

	public final Collection<Reference> register(
		Class<? extends DistributedObject> distObj, int nbInstances,
		boolean block, Reference... daemons) throws SdsException {
		objReferences.clear();
		if ((distObj == null) || (daemons == null) || (daemons.length == 0)) {
			return null;
		}

		Collection<Reference> myDaemons = new HashSet<Reference>(daemons.length);
		Collection<Reference> unreachedDaemons = new HashSet<Reference>(
			daemons.length);
		for (Reference ref : daemons) {
			if (ref != null) {
				if (ref.isInvokable()) {
					myDaemons.add(ref);
				}
				else {
					unreachedDaemons.add(ref);
				}
			}
		}

		Message msg = new Message();
		msg.setTag(CREATE_TAG_REQUEST);
		String name = distObj.getName();
		try {
			msg.setData(new ClassData(name, nbInstances, SimpleClassLoader
				.getByteCode(distObj)));
		}
		catch (IOException e) {
			throw new CommunicationException(e);
		}
		catch (Exception e) {
			throw new ActivationException(e);
		}

		unreachedDaemons.addAll(deliver(msg, block, myDaemons));
		registeredDaemons.removeAll(unreachedDaemons);
		return unreachedDaemons;
	}

	public final Collection<Reference> register(
		Class<? extends DistributedObject> distObj, int nbInstances,
		boolean block, Collection<Reference> daemons) throws SdsException {
		if ((distObj == null) || (daemons == null)) {
			return null;
		}

		Reference[] dms = new Reference[daemons.size()];
		daemons.toArray(dms);
		return register(distObj, nbInstances, block, dms);
	}

	public final Collection<Reference> killAll(boolean block) {
		Collection<Reference> tmp;
		Collection<Reference> unreachedDaemons = new HashSet<Reference>();
		Collection<Class<? extends DistributedObject>> ids = objReferences
			.keySet();
		for (Class<? extends DistributedObject> id : ids) {
			tmp = kill(id, block);
			if (tmp != null) {
				unreachedDaemons.addAll(tmp);
			}
		}

		// Clear references (as some daemons could be unreachable)
		objReferences.clear();
		return unreachedDaemons;
	}

	public final Collection<Reference> kill(
		Class<? extends DistributedObject> id, boolean block) {
		if (id == null) {
			return null;
		}

		Map<Reference, Collection<Reference>> regRefs = objReferences.get(id);
		if (regRefs == null) {
			return null;
		}

		Collection<Reference> refs;
		Collection<Reference> unreached;
		Collection<Reference> unreachedDaemons = new HashSet<Reference>(regRefs
			.size());
		Message msg;
		for (Reference issuer : regRefs.keySet()) {
			refs = regRefs.get(issuer);
			if ((refs != null) && !refs.isEmpty()) {
				msg = new Message();
				msg.setTag(KILL_TAG_REQUEST);
				msg.setData(refs);
				unreached = deliver(msg, block, issuer);
				if (unreached != null) {
					unreachedDaemons.addAll(unreached);
				}
			}
		}

		registeredDaemons.removeAll(unreachedDaemons);
		return unreachedDaemons;
	}

	public final boolean hasRegisteredDistributedObjects() {
		return !objReferences.isEmpty();
	}

	public final boolean isRegistered(Class<? extends DistributedObject> distObj) {
		return (distObj == null)
			? false
			: objReferences.containsKey(distObj);
	}

	public final Collection<Class<? extends DistributedObject>> getRegisteredDistributedObjects() {
		return objReferences.keySet();
	}

	public final Collection<Reference> getDistributedObjectReferences(
		Class<? extends DistributedObject> distObj) {
		if (distObj == null) {
			return null;
		}

		Map<Reference, Collection<Reference>> regRefs = objReferences
			.get(distObj);
		if (regRefs == null) {
			return null;
		}

		Collection<Reference> refs = new HashSet<Reference>();
		for (Reference ref : regRefs.keySet()) {
			refs.addAll(regRefs.get(ref));
		}

		return refs;
	}

	public final void clearReferences() {
		objReferences.clear();
	}

	public final boolean clearReferences(
		Class<? extends DistributedObject> distObj) {
		if (distObj == null) {
			return false;
		}

		Map<Reference, Collection<Reference>> refs = objReferences.get(distObj);
		if (refs == null) {
			return false;
		}

		objReferences.remove(distObj);
		return true;
	}
}
